---
id: api_guide
title: API
toplevel: true
---

import Figure from './figure'

This guide will cover the API level documentation for methods and tools useful
to developers looking to enhance JBrowse or write plugins.

## MenuItems objects

You can add menus or add items to existing menus in several places.

In these different places, a `MenuItem` object defines the menu item's text,
icon, action, and other attributes.

Types of `MenuItem`s:

- **Normal**: a standard menu item that performs an action when clicked
- **Checkbox**: a menu item that has a checkbox
- **Radio**: a menu item that has a radio button icon
- **Divider**: a horizontal line (not clickable) that can be used to visually
  divide menus
- **SubHeader**: text (not clickable) that can be used to visually label a
  section of a menu
- **SubMenu**: contains menu items, for making nested menus

| Name     | Description                                                                                                                                                                                              |
| -------- | -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
| type     | Options are 'normal', 'radio', 'checkbox', 'subMenu', 'subHeader', or 'divider'. If not provided, defaults to 'normal', unless a `subMenu` attribute is present, in which case it defaults to 'subMenu'. |
| label    | The text for the menu item. Not applicable to 'divider', required for all others.                                                                                                                        |
| subLabel | Additional descriptive text for the menu item. Not applicable to 'divider' or 'subHeader', optional for all others.                                                                                      |
| icon     | An icon for the menu item. Must be compatible with Material-UI's [Icons](https://material-ui.com/components/icons/). Not applicable to 'divider' or 'subHeader', optional for all others.                |
| disabled | Whether or not the menu item is disabled (meaning grayed out and not clickable). Not applicable to 'divider' or 'subHeader', optional for all others.                                                    |
| checked  | Whether or not the checkbox or radio button are selected. Only applicable to 'radio' and 'checkbox'                                                                                                      |
| onClick  | Callback of action to perform on click. Function signature is `(session) => undefined`. Required for 'normal', 'radio', and 'checkbox', not applicable to any others.                                    |
| subMenu  | An array of menu items. Applicable only to 'subMenu'.                                                                                                                                                    |

As an example, the here is an array of MenuItems and the resulting menu:

```js
;[
  {
    label: 'Normal menu item',
    icon: AddIcon,
    onClick: () => {},
  },
  {
    label: 'Normal',
    subLabel: 'with subLabel',
    icon: AddIcon,
    onClick: () => {},
  },
  {
    label: 'Disabled menu item',
    disabled: true,
    icon: AddIcon,
    onClick: () => {},
  },
  {
    type: 'radio',
    label: 'Radio checked',
    checked: true,
    onClick: () => {},
  },
  {
    type: 'radio',
    label: 'Radio unchecked',
    checked: false,
    onClick: () => {},
  },
  {
    type: 'checkbox',
    label: 'Checkbox checked',
    checked: true,
    onClick: () => {},
  },
  {
    type: 'checkbox',
    label: 'Checkbox unchecked',
    checked: false,
    onClick: () => {},
  },
  { type: 'divider' },
  { type: 'subHeader', label: 'This is a subHeader' },
  {
    label: 'SubMenu',
    subMenu: [
      {
        label: 'SubMenu item one',
        onClick: () => {},
      },
      {
        label: 'SubMenu item two',
        onClick: () => {},
      },
    ],
  },
]
```

<Figure src="/img/menu_demo.png" caption="This screenshot shows all the various track menu options, generated by the code listing"/>

## rootModel Menu API

Users can customize the top-level menu items using these functions that are
available on the rootModel:

#### `appendMenu`

Add a top-level menu

##### Parameters

| Name     | Description                 |
| -------- | --------------------------- |
| menuName | Name of the menu to insert. |

##### Return Value

The new length of the top-level menus array

#### `insertMenu`

Insert a top-level menu

##### Parameters

| Name     | Description                                                                                                                                 |
| -------- | ------------------------------------------------------------------------------------------------------------------------------------------- |
| menuName | Name of the menu to insert.                                                                                                                 |
| position | Position to insert menu. If negative, counts from the end, e.g. `insertMenu('My Menu', -1)` will insert the menu as the second-to-last one. |

##### Return Value

The new length of the top-level menus array

#### `appendToMenu`

Add a menu item to a top-level menu

##### Parameters

| Name     | Description                              |
| -------- | ---------------------------------------- |
| menuName | Name of the top-level menu to append to. |
| menuItem | Menu item to append.                     |

##### Return Value

The new length of the menu

#### `insertInMenu`

Insert a menu item into a top-level menu

##### Parameters

| Name     | Description                                                                                                                                      |
| -------- | ------------------------------------------------------------------------------------------------------------------------------------------------ |
| menuName | Name of the top-level menu to insert into.                                                                                                       |
| menuItem | Menu item to insert.                                                                                                                             |
| position | Position to insert menu item. If negative, counts from the end, e.g. `insertMenu('My Menu', -1)` will insert the menu as the second-to-last one. |

##### Return Value

The new length of the menu

#### `appendToSubMenu`

Add a menu item to a sub-menu

##### Parameters

| Name     | Description                                                                                   |
| -------- | --------------------------------------------------------------------------------------------- |
| menuPath | Path to the sub-menu to add to, starting with the top-level menu (e.g. `['File', 'Insert']`). |
| menuItem | Menu item to append.                                                                          |

##### Return value

The new length of the sub-menu

#### `insertInSubMenu`

Insert a menu item into a sub-menu

##### Parameters

| Name     | Description                                                                                                                                      |
| -------- | ------------------------------------------------------------------------------------------------------------------------------------------------ |
| menuPath | Path to the sub-menu to add to, starting with the top-level menu (e.g. `['File', 'Insert']`).                                                    |
| menuItem | Menu item to insert.                                                                                                                             |
| position | Position to insert menu item. If negative, counts from the end, e.g. `insertMenu('My Menu', -1)` will insert the menu as the second-to-last one. |

##### Return value

The new length of the sub-menu
